/* gfunctions.c - simple generic graphic functions used for plotting
 * "Alpha" state. Use at your own risk!
 * Copyright 2012 Mihai Gaitos, mihaig@hawk.ro
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN
 * IF THE COPYRIGHT HOLDER HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 * THE COPYRIGHT HOLDER SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE COPYRIGHT HOLDER HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */
#include "gfunctions.h"
#ifdef _SDL_H
#include "sdlplot.c"
#else
#include "mplot.c"
#endif
static int setpixelv(pixel_t * pixel,unsigned int red,unsigned int green,unsigned int blue)
{
if(red>255) red=255;
if(green>255) green=255;
if(blue>255) blue=255;

	
pixel->r=red;
pixel->g=green;
pixel->b=blue;
}


int rect(void *screen,int x1, int y1, int x2, int y2, pixel_t *pixel, int mod)

{
int x,y;
if((mod&R_FILLED)!=0)
	{
	for(x=x1;x<=x2;x++)
		for(y=y1;y<=y2;y++)
			plot(screen,x,y,pixel,mod);
	}
else
	{
	draw(screen,x1,y1,x2,y1,pixel,mod);
	draw(screen,x2,y1,x2,y2,pixel,mod);
	draw(screen,x2,y2,x1,y2,pixel,mod);
	draw(screen,x1,y2,x1,y1,pixel,mod);
	}
return 1;
}

int draw(void *screen, int x, int y, int x1, int y1, pixel_t *pixel, int mod)
{
int e2;
int dx=abs(x1-x);
int dy=abs(y1-y);
int sx,sy;
unsigned int maxx,maxy;
getsurfacesize(screen,&maxx,&maxy);
if(x<x1)
	sx=1;
else
	sx=-1;
if(y<y1)
	sy=1;
else
	sy=-1;
if(x==x1)
	sx=0;
if(y==y1)
	sy=0;


int err=dx-dy;
while( (x!=x1 || y!=y1) && x<maxx && y<maxy && x>=0 && y>=0)
	{
	plot(screen,x,y,pixel,mod);
	e2=err*2;
	if(e2>-dy)
		{
		err=err-dy;
		x=x+sx;
		}
	if(e2<dx)
		{
		err=err+dx;
		y=y+sy;
		}
	
	}
}


unsigned char *chardata(font_t *f,int c)
{
return(f->data+c*f->height*f->wbytes);
}

int drawstr(void *screen, font_t *font,int x, int y, unsigned char *t, pixel_t *px, pixel_t *bg)
{
unsigned int cx,i;
unsigned int mx,my;
getsurfacesize(screen,&mx,&my);
cx=x;
unsigned int l=strlen(t);
for(i=0;(i<l)&&(cx<mx);i++)
	{
	drawchar(screen,font,cx,y,t[i],px,bg);
	cx+=font->width;
	}
}


int drawchar(void *screen, font_t *font,int x, int y, int c, pixel_t *pixel, pixel_t *bg)
{
int px,py,pb,cfp;
unsigned int cl;
unsigned char *cd=chardata(font,c);
for(py=0;py<font->height;py++)	// font line
	{
	pb=0;
	for(px=0;px<font->width;px++) // font bit
		{
		if((px&7)==0)
			{
			cl=cd[py*font->wbytes+pb];
			pb++;
			}
		if((cl&1)==1)
			plot(screen,x+px,y+py,pixel,0);
		else
			if(bg!=NULL) plot(screen,x+px,y+py,bg,0);
		cl=cl>>1;
		}
	}
}

font_t *loadfont(unsigned char *name)
{
font_t *f=malloc(sizeof(font_t));
int fd=open(name,O_RDONLY);
if(fd<0)
	{
	fprintf(stderr,"Font file not found.\n");
	free(f);
	return NULL;
	}
int res=read(fd,f,sizeof(font_t));
if(res<sizeof(font_t))
	{
	fprintf(stderr,"Error reading font header.\n");
	free(f);
	return NULL;
	}
f=realloc(f,f->size);
res=read(fd,&f->data,f->size-sizeof(font_t));
if(res<(f->size-sizeof(font_t)))
	{
	fprintf(stderr,"Error reading font data. Read %d, expected %u.\n",res,f->size-sizeof(font_t));
	free(f);
	return NULL;
	}
close(fd);
return f;
}

