/* imdec.c - Aurora .im file decoder
 * Copyright 2015 Mihai Gaitos, mihaig@hawk.ro
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN
 * IF THE COPYRIGHT HOLDER HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 * THE COPYRIGHT HOLDER SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE COPYRIGHT HOLDER HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

#include "imstructs.h" /* Aurora image file structures */

typedef struct
{
	uint8_t	r;
	uint8_t g;
	uint8_t b;
} palentry; /* this is not part of Aurora image file structure; created here for convenience */

/* fill picture buffer tb with values from src+s. cx,cy = start coordinates, sx,sy = rectangle size
 * pw, ph = picture width and height */
int fillpb(void *src,unsigned int s,unsigned char *tb,int cx,int cy,int sx,int sy,int pw,int ph)
{
rtb *rt=src+s;
unsigned int x,y;
uint8_t v;
int r;
r=rt->run+1; v=rt->val;
int bpos=(ph-cy-1)*pw+cx;

unsigned long xa;
xa=((void *)rt-src);

for(y=0;y<sy;y++)
	{
	for(x=0;x<sx;x++)
		{
		tb[bpos+x]=v; r--;
		if(r==0)
			{
			rt++;
			r=rt->run+1;
			v=rt->val;
			}
		}
	bpos-=pw;
	}
}

/* process rtah image data at buf+of, write result into tb (malloced here) also sets width and height */
void procrlea(void *buf,unsigned int of, unsigned char **tb, unsigned int *pw, unsigned int *ph)
{
int i,f;
unsigned int pos,rtpos;
unsigned char *cb=buf;
unsigned int cx=0,cy=0,sx,sy;
rtah *rth=buf+of;
*pw=rth->s1;
*ph=rth->s2;
/* printf("\nAt %d (0x%X) a4=%d, S1=%d, S2=%d\n",of,of,rth->a4,rth->s1,rth->s2); */
*tb=malloc(rth->s1*rth->s2);
pos=rth->rtabegh<<16|rth->rtabegl;
rta *rtarr=buf+pos;
i=0;
/* int ateoi=0; */
for(i=0;rtarr[i].t1==7;i++)
	{
/*	if(ateoi) { printf("."); continue; } */
	rtpos=(rtarr[i].offh<<16)|(rtarr[i].offl);
	if(*pw-cx<16) sx=*pw-cx; else sx=16;
	if(*ph-cy<16) sy=*ph-cy; else sy=16; /* this is not needed at every step but it's clearer like this */
	fillpb(buf,rtpos,*tb,cx,cy,sx,sy,*pw,*ph);
/*	printf("At %d,%d with size %d.%d\n",cx,cy,sx,sy); */
	if(cx+sx<*pw) cx+=sx; else
		{
		if(cy+sy>=*ph) break; /* { printf("End of image at rec. no. %d",i); ateoi=1; continue; } */
		cx=0;
		cy+=sy;
		}
	}
/* printf(" found %d\n",i); */
}

/* process palette - from Aurora format to palette record format */
/* this mallocs memory for a whole (256 colors) palette */
procpal(void *buf,unsigned int of,palentry **pp)
{
palh *ph=buf+of;
char *ps=buf+of+4;
palentry *mp=malloc(256*sizeof(palentry));
int i;
for(i=0;i<ph->ccnt;i++)
	{
	mp[i].r=ps[i];
	mp[i].g=ps[i+128];
	mp[i].b=ps[i+256];
	}
*pp=mp;
}

/* process contents of .im file from buf, write pixels to tb, palette to pal, width and height to sx and sy */
/* limited to first image (type 0x10) section */
int procim(void *buf,unsigned char **tb,palentry **pal,unsigned int *sx, unsigned int *sy)
{
imheada *imh;
imh=buf+8;
int i=0;
while(imh[i].t1!=0)
	{
	if(imh[i].t1==0x20) procpal(buf,imh[i].offh<<16|imh[i].offl,pal);
	if(imh[i].t1==0x10) { procrlea(buf,imh[i].offh<<16|imh[i].offl,tb,sx,sy); break; } /* only show first image section */
	i++;
	}
}

/* "show" picture from ibuf on screen. This is a bit of a misnomer, in fact it simply converts from an 8bpp linear
 * image buffer to the screen specific format, be it SDL_Surface or M_Surface (generic/memory; see mplot.c)
 * this could be optimized further, as it's based on an experimental function to check various types of raw
 * image buffers.
 * it will shift the displayed image on the screen by xpos and ypos.
 * if pal is NULL it will decode the byte as IIRRGGBB.
 */
void showpic(void *screen,unsigned char *ibuf,palentry *pal,unsigned int sx,unsigned int sy,unsigned int xpos,unsigned int ypos)
{
unsigned int x,y;
pixel_t mp;
unsigned int cbp=0;
unsigned char c;
for(y=0;y<sy;y++)
	{
	for(x=0;x<sx;x++)
		{
		c=ibuf[cbp];
		cbp++;
		if(pal!=NULL)
			{
			mp.r=pal[c].r;
			mp.g=pal[c].g;
			mp.b=pal[c].b;
			}
		else /* in case of no palette, use some generic values */
			{
			mp.r=(c&0x30);
			mp.g=(c&0x0c)<<2;
			mp.b=(c&0x03)<<4;
			c=c&0xc0;
			mp.r|=c;
			mp.g|=c;
			mp.b|=c;
			}
		plot(screen,xpos+x,ypos+y,&mp,0);
		}
	}
}

/* read .im file into memory and check for header signature. will return 0 on success */
int readim(char *fn,void **ibuf)
{
struct stat finfos;
/* read file into memory */
int fdi=open(fn,O_RDONLY);
if(fdi<1)
	{
	fprintf(stderr,"Unable to open %s\n",fn);
	return -2;
	}
if(fstat(fdi,&finfos)!=0)
	{
	fprintf(stderr,"Unable to stat %s\n",fn);
	close(fdi);
	return -3;
	}
off_t isize=finfos.st_size;
if(isize>20*1024*1024)	/* 20MB should be sane enough to avoid accidental loading of huge files */
	{
	fprintf(stderr,"File too big!\n");
	close(fdi);
	return 1;
	}
*ibuf=malloc(isize);
if(read(fdi,*ibuf,isize)!=isize)
	{
	fprintf(stderr,"Unable to read %s\n",fn);
	close(fdi);
	return -4;
	}
close(fdi);
uint16_t sgn=*(uint16_t *) *ibuf;
if(sgn!=AURSIG)
	{
	fprintf(stderr,"Invalid signature. Found 0x%04x, expecting 0x%04x\n",sgn,AURSIG);
	return 2;
	}
return 0;
}
