/*
 * Copyright 2002-2010 Guillaume Cottenceau.
 *
 * This software may be freely redistributed under the terms
 * of the X11 license.
 *
 */


/*
 * Modified by Mihai Gaitos in 2015, for use as "png exporter"
 * only supports 24bit RGB
 * to be used with gfunctions.c and mplot.c
 */

#ifndef PNG_H
#include <png.h>
#endif

void abort_(const char * s, ...)
{
        va_list args;
        va_start(args, s);
        vfprintf(stderr, s, args);
        fprintf(stderr, "\n");
        va_end(args);
        abort();
}



void save_png_to_file(void *scr,char* file_name)
{
	int width, height;
	pixel_t cp;				/* defined in gfunctions.h */
	getsurfacesize(scr,&width,&height);

	png_byte color_type=PNG_COLOR_TYPE_RGB;	/* HARDCODED - the only one supported/used */
	png_byte psize=3;			/* HARDCODED and related to above! */
	png_byte bit_depth=8;			/* HARDCODED and related to above */
	png_structp png_ptr;
	png_infop info_ptr;
	int number_of_passes;
	png_bytep * row_pointers;
	png_byte * row;
	int x, y;


        /* create file */
        FILE *fp = fopen(file_name, "wb");
        if (!fp)
                abort_("[write_png_file] File %s could not be opened for writing", file_name);


        /* initialize stuff */
        png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);

        if (!png_ptr)
                abort_("[write_png_file] png_create_write_struct failed");

        info_ptr = png_create_info_struct(png_ptr);
        if (!info_ptr)
                abort_("[write_png_file] png_create_info_struct failed");

        if (setjmp(png_jmpbuf(png_ptr)))
                abort_("[write_png_file] Error during init_io");

        png_init_io(png_ptr, fp);


	/* create rows */
	row_pointers=png_malloc(png_ptr,height*sizeof(png_byte *));
	for(y=0;y<height;y++)
		{
		row=png_malloc(png_ptr,width*psize);
		row_pointers[y]=row;
		for(x=0;x<width;x++)
			{
			getpixel(scr, x, y,&cp);
			*row++ = cp.r; *row++ = cp.g; *row++ = cp.b;
			}
		}


        /* write header */
        if (setjmp(png_jmpbuf(png_ptr)))
                abort_("[write_png_file] Error during writing header");

        png_set_IHDR(png_ptr, info_ptr, width, height,
                     bit_depth, color_type, PNG_INTERLACE_NONE,
                     PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);

        png_write_info(png_ptr, info_ptr);


        /* write bytes */
        if (setjmp(png_jmpbuf(png_ptr)))
                abort_("[write_png_file] Error during writing bytes");

        png_write_image(png_ptr, row_pointers);


        /* end write */
        if (setjmp(png_jmpbuf(png_ptr)))
                abort_("[write_png_file] Error during end of write");

        png_write_end(png_ptr, NULL);

        /* cleanup heap allocation */
        for (y=0; y<height; y++)
                free(row_pointers[y]);
        free(row_pointers);

        fclose(fp);
}


